/*
 *  TKnobControl.cpp
 *  NeuSynth
 *
 *  Created by Airy Andr on Sun Oct 20 2002.
 *  Copyright (c) 2002 __MyCompanyName__. All rights reserved.
 *
 */

#include "TKnobControl.h"
#include "TImageCache.h"

// -----------------------------------------------------------------------------
//	prototypes
// -----------------------------------------------------------------------------
//
static void HIRectToQDRect(
		    const HIRect*		inHIRect,
		    Rect*				outQDRect );

// -----------------------------------------------------------------------------
//	macros
// -----------------------------------------------------------------------------
//

CFBundleRef TKnobControl::sStaticBundle = 0;
// -----------------------------------------------------------------------------
//	TKnobControl constructor
// -----------------------------------------------------------------------------
//
TKnobControl::TKnobControl(
				 HIViewRef			inControl )
:	TViewNoCompositingCompatible( inControl ), mImage(0), mBackImage(0),
        mXOffset(0), mYOffset(0), mMinAngle(-135), mMaxAngle(135)
{
    ChangeAutoInvalidateFlags( kAutoInvalidateOnActivate | kAutoInvalidateOnEnable, 0 );
    mBundleRef = sStaticBundle;
    if (mBundleRef)
	CFRetain(mBundleRef);
}

// -----------------------------------------------------------------------------
//	TKnobControl destructor
// -----------------------------------------------------------------------------
//	Clean up after yourself.
//
TKnobControl::~TKnobControl()
{
    if (mImage)
	CGImageRelease(mImage);
    if (mBackImage)
	CGImageRelease(mBackImage);
    if (mBundleRef)
	CFRelease(mBundleRef);
}

// -----------------------------------------------------------------------------
//	SetClassBundle
// -----------------------------------------------------------------------------
//
void TKnobControl::SetClassBundle(CFBundleRef sBundle)
{
    if (sStaticBundle)
	CFRelease(sBundle);

    if (sBundle)
	CFRetain (sBundle);
    sStaticBundle = sBundle;
}

// -----------------------------------------------------------------------------
//	GetKind
// -----------------------------------------------------------------------------
//
ControlKind TKnobControl::GetKind()
{
    const ControlKind kMyKind = { 'airy', 'knob' };

    return kMyKind;
}

//-----------------------------------------------------------------------------------
//	Create
//-----------------------------------------------------------------------------------
//
OSStatus TKnobControl::Create(
				   HIViewRef*			outControl,
				   const HIRect*		inBounds,
				   WindowRef			inWindow,
				   CFStringRef 	classID)
{
    OSStatus			err;
    EventRef			event = CreateInitializationEvent();
    ControlRef			root;

    RegisterClass(classID);
    err = HIObjectCreate( classID, event, (HIObjectRef*) outControl );

    ReleaseEvent( event );

    require_noerr( err, CantCreateHIObject );

    if ( inWindow != NULL )
    {
	GetRootControl( inWindow, &root );
	HIViewAddSubview( root, *outControl );
    }

    if ( inBounds != NULL )
	HIViewSetFrame( *outControl, inBounds );

CantCreateHIObject:

	return err;
}

//-----------------------------------------------------------------------------------
//	RegisterClassForBundleID
//-----------------------------------------------------------------------------------
//
void TKnobControl::RegisterClassForBundleID(CFStringRef str)
{
    CFMutableStringRef reg = CFStringCreateMutableCopy(NULL, 100, str);
    CFStringAppend(reg, CFSTR(".knob"));
TView::RegisterSubclass( reg, Construct);
    CFRelease(reg);
}
//-----------------------------------------------------------------------------------
//	RegisterClass
//-----------------------------------------------------------------------------------
//
void TKnobControl::RegisterClass(CFStringRef str)
{
TView::RegisterSubclass( str, Construct);
}
//-----------------------------------------------------------------------------------
//	Construct
//-----------------------------------------------------------------------------------
//
OSStatus TKnobControl::Construct(
				    ControlRef		inControl,
				    TView**			outView )
{
    *outView = new TKnobControl( inControl );

    return noErr;
}

//-----------------------------------------------------------------------------------
//	Initialize
//-----------------------------------------------------------------------------------
//	The control is set up.  Do the last minute stuff that needs to be done like
//	installing EventLoopTimers.
//
OSStatus TKnobControl::Initialize(
				     TCarbonEvent&		inEvent )
{
#pragma unused( inEvent )
    TitleChanged();
    return noErr;
}
//-----------------------------------------------------------------------------------
//	ValueChanged
//-----------------------------------------------------------------------------------
//
OSStatus TKnobControl::ValueChanged()
{
    Invalidate();
    return noErr;
}
//-----------------------------------------------------------------------------------
//	TitleChanged
//	The title format is
//	 	filename minAngle maxAngle xOffset xOffset
//	With :
//		<filename>.pgn and <filename>Back.png are the pict used by the control
//		minAngle and maxAngle the angle for min and max values (-135,135 default values)
//		xOffset and yOffset the offset of the rotating part from the back (0,0 default values)
//		Only the filename is mandatory
//-----------------------------------------------------------------------------------
OSStatus TKnobControl::TitleChanged()
{
    if (mImage)
	CGImageRelease(mImage);
    if (mBackImage)
	CGImageRelease(mBackImage);
    if ( mBundleRef )
    {
	CFStringRef fileName;
	CopyControlTitleAsCFString(GetViewRef(), &fileName);
	if ( fileName != NULL )
	{
	    char buffer[100];
	    char name[100];

	    char file[100];
	    float xOffset, yOffset, minAngle, maxAngle;

	    CFStringGetCString(fileName, buffer, 100, kCFStringEncodingASCII);
	    int nbVal = sscanf(buffer, "%s %f %f %f %f", file, &minAngle, &maxAngle, &xOffset, &yOffset);
	    if (nbVal < 2) minAngle = -135.;
	    if (nbVal < 3) maxAngle = 135.;
	    if (nbVal < 4) xOffset = 0.;
	    if (nbVal < 5) yOffset = 0.;
	    mMinAngle = minAngle;
	    mMaxAngle = maxAngle;
	    mXOffset = xOffset;
	    mYOffset = yOffset;
	    
	    sprintf(name, "%s.png", file);
	    CFStringRef knob = CFStringCreateWithCStringNoCopy(0, name, kCFStringEncodingASCII, 0);

	    CFURLRef url = ::CFBundleCopyResourceURL( mBundleRef, knob, NULL, NULL );
	    if (url) {
		mImage = TImageCache::GetImage(url);
		CFRelease( url );
	    }

	    sprintf(name, "%sBack.png", file);
	    CFStringRef back = CFStringCreateWithCStringNoCopy(0, name, kCFStringEncodingASCII, 0);
	    url = ::CFBundleCopyResourceURL( mBundleRef, back, NULL, NULL );
	    if (url) {
		mBackImage = TImageCache::GetImage(url);
		CFRelease( url );
	    }
	    CFRelease( fileName );
	}
    }
    return noErr;
}

//-----------------------------------------------------------------------------------
//	Draw
//-----------------------------------------------------------------------------------
//	The fun part of the control
//
void TKnobControl::CompatibleDraw(
			   RgnHandle				inLimitRgn,
			   CGContextRef			inContext,
				  bool  inCompositing)
{
#pragma unused( inLimitRgn )
    TRect					bounds = Bounds();
    CGContextRef			context = inContext;

    TRect knobBounds = bounds;
    
    if (mBackImage)
	HIViewDrawCGImage( context, &bounds, mBackImage);

    if (mImage) {
	float angle = (((float) (GetValue() - GetMinimum())/ (float) (GetMaximum()-GetMinimum())) * (mMaxAngle-mMinAngle) + mMinAngle) / 360.f;
	if (mBackImage)
	{
	    float scaleX = float(CGImageGetWidth(mImage)) / float(CGImageGetWidth(mBackImage));
	    float scaleY = float(CGImageGetHeight(mImage)) / float(CGImageGetHeight(mBackImage));
	    knobBounds.SetWidth(knobBounds.Width() * scaleX);
	    knobBounds.SetHeight(knobBounds.Height() * scaleY);
	}
	float knobWidth = knobBounds.Width();
	float knobHeight = knobBounds.Height();

	CGContextTranslateCTM(context, mXOffset + (bounds.MinX() + bounds.MaxX())/2., mYOffset + (bounds.MinY() + bounds.MaxY())/2.);
	CGContextRotateCTM( context, 2.0 * M_PI * angle );
	knobBounds.MoveTo(-knobWidth/2, -knobHeight/2);
	HIViewDrawCGImage( context, &knobBounds, mImage);
    }
}
// -----------------------------------------------------------------------------
//	HitTest
// -----------------------------------------------------------------------------
//	Asks your view to return what part of itself (if any) is hit by the point
//	given to it. The point is in VIEW coordinates, so you should get the view
//	rect to do bounds checking.
//
ControlPartCode TKnobControl::HitTest(
				     const HIPoint&		inWhere )
{
    ControlPartCode		part;

    // is the mouse on the button?
    if( CGRectContainsPoint( Bounds(), inWhere ) )
	part = kControlButtonPart;
    else
	part = kControlNoPart;

    return part;
}
// -----------------------------------------------------------------------------
//	Track
// -----------------------------------------------------------------------------
//	Default tracking method.
//
OSStatus TKnobControl::Track(
			    TCarbonEvent&		inEvent,
			    ControlPartCode*	outPart )
{
    Point pt, org;
    float deltaX, deltaY;
    int stillTracking = 1;
    SInt32 mini, maxi;

    mini = GetMinimum();
    maxi = GetMaximum();
    
    GetMouse ( &org );
    while (stillTracking)
    {
	MouseTrackingResult outResult;
	
	TrackMouseLocation (NULL, &pt, &outResult);
	if ( outResult == kMouseTrackingMouseUp ) {
	    stillTracking = 0;
	} else {
	    deltaX = (float) (pt.h - org.h);
	    deltaY = (float) (pt.v - org.v);

	    SInt32 curVal = GetValue();
	    SInt32 val = SInt32(rint(curVal+(deltaX-deltaY)*(maxi-mini)/100.));
	    if ( val > maxi) val = maxi;
	    if ( val < mini) val = mini;
	    if (val != curVal) {
		SetValue ( val );
		org = pt;
	    }
	}
    }
    *outPart = kControlNoPart;
    return noErr;
}

//-----------------------------------------------------------------------------------
//	GetRegion
//-----------------------------------------------------------------------------------
//
OSStatus TKnobControl::GetRegion(
				    ControlPartCode		inPart,
				    RgnHandle			outRgn )
{
    OSStatus			err = eventNotHandledErr;
    HIRect				bounds;
    Rect				qdBounds;

    switch ( inPart )
    {
	case kControlContentMetaPart:
	case kControlStructureMetaPart:
	    bounds = Bounds();
	    HIRectToQDRect( &bounds, &qdBounds );
	    RectRgn( outRgn, &qdBounds );
	    err = noErr;
	    break;
    }

    return err;
}


//-----------------------------------------------------------------------------------
//	HIRectToQDRect
//-----------------------------------------------------------------------------------
//	Converts an HIRect to a Quickdraw rectangle, snipping off non-integral bits.
//
static void HIRectToQDRect(
		    const HIRect*	inHIRect,
		    Rect*			outQDRect )
{
    outQDRect->top = (SInt16) CGRectGetMinY( *inHIRect );
    outQDRect->left = (SInt16) CGRectGetMinX( *inHIRect );
    outQDRect->bottom = (SInt16) CGRectGetMaxY( *inHIRect );
    outQDRect->right = (SInt16) CGRectGetMaxX( *inHIRect );
}


